from floe import make_var, FP
from Parameter import Parameter


import uasyncio as asyncio

from utime import sleep_us, time
from machine import Pin
from micropython import const


class HX711Exception(Exception):
    pass


class InvalidMode(HX711Exception):
    pass


class DeviceIsNotReady(HX711Exception):
    pass


class HX711Core:
    """
    Micropython driver for Avia Semiconductor's HX711
    24-Bit Analog-to-Digital Converter
    """
    CHANNEL_A_128 = const(1)
    CHANNEL_A_64 = const(3)
    CHANNEL_B_32 = const(2)

    DATA_BITS = const(24)
    MAX_VALUE = const(0x7fffff)
    MIN_VALUE = const(0x800000)
    READY_TIMEOUT_SEC = const(5)
    SLEEP_DELAY_USEC = const(80)

    def __init__(self, d_out: int, pd_sck: int, channel: int = CHANNEL_A_128):
        self.d_out_pin = Pin(d_out, Pin.IN)
        self.pd_sck_pin = Pin(pd_sck, Pin.OUT, value=0)
        self.channel = channel
        self.zero = 0
        self.state = 0
        self.old = 0

    def __repr__(self):
        return "HX711 on channel %s, gain=%s" % self.channel

    def _convert_from_twos_complement(self, value: int) -> int:
        """
        Converts a given integer from the two's complement format.
        """
        if value & (1 << (self.DATA_BITS - 1)):
            value -= 1 << self.DATA_BITS
        return value

    def _set_channel(self):
        """
        Input and gain selection is controlled by the
        number of the input PD_SCK pulses
        3 pulses for Channel A with gain 64
        2 pulses for Channel B with gain 32
        1 pulse for Channel A with gain 128
        """
        for i in range(self._channel):
            self.pd_sck_pin.value(1)
            self.pd_sck_pin.value(0)

    def _wait(self):
        """
        If the HX711 is not ready within READY_TIMEOUT_SEC
        the DeviceIsNotReady exception will be thrown.
        """
        t0 = time()
        while not self.is_ready():
            if time() - t0 > self.READY_TIMEOUT_SEC:
                raise DeviceIsNotReady()

    @property
    def channel(self) -> tuple:
        """
        Get current input channel in a form
        of a tuple (Channel, Gain)
        """
        if self._channel == self.CHANNEL_A_128:
            return 'A', 128
        if self._channel == self.CHANNEL_A_64:
            return 'A', 64
        if self._channel == self.CHANNEL_B_32:
            return 'B', 32

    @channel.setter
    def channel(self, value):
        """
        Set input channel
        HX711.CHANNEL_A_128 - Channel A with gain 128
        HX711.CHANNEL_A_64 - Channel A with gain 64
        HX711.CHANNEL_B_32 - Channel B with gain 32
        """
        if value not in (self.CHANNEL_A_128, self.CHANNEL_A_64, self.CHANNEL_B_32):
            raise InvalidMode('Gain should be one of HX711.CHANNEL_A_128, HX711.CHANNEL_A_64, HX711.CHANNEL_B_32')
        else:
            self._channel = value

        if not self.is_ready():
            self._wait()

        for i in range(self.DATA_BITS):
            self.pd_sck_pin.value(1)
            self.pd_sck_pin.value(0)

        self._set_channel()

    def is_ready(self) -> bool:
        """
        When output data is not ready for retrieval,
        digital output pin DOUT is high.
        """
        return self.d_out_pin.value() == 0

    def power_off(self):
        """
        When PD_SCK pin changes from low to high
        and stays at high for longer than 60 us ,
        HX711 enters power down mode.
        """
        self.pd_sck_pin.value(0)
        self.pd_sck_pin.value(1)
        sleep_us(self.SLEEP_DELAY_USEC)

    def power_on(self):
        """
        When PD_SCK returns to low, HX711 will reset
        and enter normal operation mode.
        """
        self.pd_sck_pin.value(0)
        self.channel = self._channel

    def tare(self):
        self.zero = self.read(corrected=False)
        print(f'taring, val = {self.zero}')

    # def check(self):
    #     self.state = self.read() - self.zero
    #     # print('{}, {}'.format(self.state, self.old))
    #     corrected = self.state / 9000
    #     if corrected > 255:
    #         corrected = 255
    #     elif corrected < 0:
    #         corrected = 0
    #     # print(corrected)
    #     corrected = int(corrected)
    #     if abs(corrected - self.old) > 1:
    #         # print(corrected)
    #         self.old = corrected
    #     # if abs(self.state - self.old) > 1:
    #     #     print(self.state - self.zero)
    #     #     self.old = self.state
    #         return corrected

    def read(self, corrected=True):
        val = self._read()
        if corrected:
            return val - self.zero
        else:
            return val
        
    def _read(self, raw=False):
        """
        Read current value for current channel with current gain.
        if raw is True, the HX711 output will not be converted
        from two's complement format.
        """
        if not self.is_ready():
            self._wait()  # TODO: make this async fucntion

        raw_data = 0
        for i in range(self.DATA_BITS):
            self.pd_sck_pin.value(1)
            self.pd_sck_pin.value(0)
            raw_data = raw_data << 1 | self.d_out_pin.value()
        self._set_channel()

        if raw:
            return raw_data
        else:
            return self._convert_from_twos_complement(raw_data)


class Hx711(Parameter):
    struct = 'f'
    channels = {
        "CHANNEL_A_128": HX711Core.CHANNEL_A_128,
        "CHANNEL_A_64": HX711Core.CHANNEL_A_64,
        "CHANNEL_B_32": HX711Core.CHANNEL_B_32
    }
    def __init__(self, *, tare, d_out, pd_sck, channel, **k):
        super().__init__(**k)
        self.tare = make_var(tare)
        
        _channel = self.channels[channel]
        self.scale = HX711Core(d_out=d_out, pd_sck=pd_sck, channel=_channel)
        
        self.ring = [0]
        self.index  = 0
        
        loop = asyncio.get_event_loop()
        loop.create_task(self.chk())
        
    def update(self):
        super().update()
        self.tare.add_hot(self.do_tare)
        self.scale.tare()
        sleep_us(500)
        self.state = self.scale.read()
        self.ring[0] = self.state
        self.ring = self.ring*3
        
    
    def do_tare(self, state):
        if state == True:
            self.scale.tare()
    
    async def chk(self) -> None:
        while True:    
            self.ring[self.index] = self.scale.read()
            self.index += 1
            if self.index == len(self.ring):
                self.index = 0
            state = sum(self.ring)//len(self.ring)
            
            if abs(self.state - state) > 200:
                self.state = state
                self.send()
            await asyncio.sleep_ms(40)

    

